<h2 id="puppetdb">PuppetDB {{ page.doc.my_versions.puppetdb }}</h2>

{% md %}

* **General information**
    * [Overview and requirements]({{puppetdb}}/index.html)
    * [Contributing to PuppetDB]({{puppetdb}}/CONTRIBUTING.html)
    * [Frequently asked questions]({{puppetdb}}/puppetdb-faq.html)
    * [Release notes]({{puppetdb}}/release_notes.html)
    * [Versioning policy]({{puppetdb}}/versioning_policy.html)
    * [Known issues]({{puppetdb}}/known_issues.html)
    * [Community add-ons]({{puppetdb}}/community_add_ons.html)
* **Installation**
    * [Installing via Puppet module]({{puppetdb}}/install_via_module.html)
    * [Installing from packages]({{puppetdb}}/install_from_packages.html)
    * [Installing from source]({{puppetdb}}/install_from_source.html)
    * [Upgrading PuppetDB]({{puppetdb}}/upgrade.html)
    * [Connecting Puppet masters]({{puppetdb}}/connect_puppet_master.html)
    * [Connecting standalone Puppet nodes]({{puppetdb}}/connect_puppet_apply.html)
* **Configuration**
    * [Configuring PuppetDB]({{puppetdb}}/configure.html)
    * [puppetdb.conf: Configuring a Puppet/PuppetDB connection]({{puppetdb}}/puppetdb_connection.html)
    * [Setting up SSL for PostgreSQL]({{puppetdb}}/postgres_ssl.html)
* **Usage/admin**
    * [Using PuppetDB]({{puppetdb}}/using.html)
    * [Maintaining and tuning]({{puppetdb}}/maintain_and_tune.html)
    * [PuppetDB CLI]({{puppetdb}}/pdb_client_tools.html)
    * [Exporting and anonymizing data]({{puppetdb}}/anonymization.html)
    * [Scaling recommendations]({{puppetdb}}/scaling_recommendations.html)
    * [Debugging with remote REPL]({{puppetdb}}/repl.html)
    * [Load testing]({{puppetdb}}/load_testing_tool.html)
* **Troubleshooting**
    * [General Support Guide]({{puppetdb}}/pdb_support_guide.html)
    * [Session logging]({{puppetdb}}/trouble_session_logging.html)
* **PQL - Puppet Query Language**
    * [Tutorial]({{puppetdb}}/api/query/tutorial-pql.html)
    * [Reference guide]({{puppetdb}}/api/query/v4/pql.html)
    * [Examples]({{puppetdb}}/api/query/examples-pql.html)

* **API**
    * [Overview]({{puppetdb}}/api/index.html)
    * [Query tutorial]({{puppetdb}}/api/query/tutorial.html)
    * [Curl tips]({{puppetdb}}/api/query/curl.html)
* **Query API version 4**
    * [Upgrading from version 3]({{puppetdb}}/api/query/v4/upgrading-from-v3.html)
    * [Query structure]({{puppetdb}}/api/query/v4/query.html)
    * [Entities]({{puppetdb}}/api/query/v4/entities.html)
    * [AST query language]({{puppetdb}}/api/query/v4/ast.html)
    * [Query paging]({{puppetdb}}/api/query/v4/paging.html)
    * [Root endpoint]({{puppetdb}}/api/query/v4/index.html)
    * [Nodes endpoint]({{puppetdb}}/api/query/v4/nodes.html)
    * [Environments endpoint]({{puppetdb}}/api/query/v4/environments.html)
    * [Producers endpoint]({{puppetdb}}/api/query/v4/producers.html)
    * [Factsets endpoint]({{puppetdb}}/api/query/v4/factsets.html)
    * [Facts endpoint]({{puppetdb}}/api/query/v4/facts.html)
    * [Fact-names endpoint]({{puppetdb}}/api/query/v4/fact-names.html)
    * [Fact-paths endpoint]({{puppetdb}}/api/query/v4/fact-paths.html)
    * [Fact-contents endpoint]({{puppetdb}}/api/query/v4/fact-contents.html)
    * [Inventory endpoint]({{puppetdb}}/api/query/v4/inventory.html)
    * [Catalogs endpoint]({{puppetdb}}/api/query/v4/catalogs.html)
    * [Resources endpoint]({{puppetdb}}/api/query/v4/resources.html)
    * [Edges endpoint]({{puppetdb}}/api/query/v4/edges.html)
    * [Reports endpoint]({{puppetdb}}/api/query/v4/reports.html)
    * [Events endpoint]({{puppetdb}}/api/query/v4/events.html)
    * [Event counts endpoint]({{puppetdb}}/api/query/v4/event-counts.html)
    * [Aggregate event counts endpoint]({{puppetdb}}/api/query/v4/aggregate-event-counts.html)
    * [Package endpoints]({{puppetdb}}/api/query/v4/packages.html)
* **Extensions API version 1 (PE-only)**
    * [State overview endpoint]({{puppetdb}}/api/ext/v1/state-overview.html)
* **Admin API version 1**
    * [Archive endpoint]({{puppetdb}}/api/admin/v1/archive.html)
    * [Command (cmd) endpoint]({{puppetdb}}/api/admin/v1/cmd.html)
    * [Summary stats endpoint]({{puppetdb}}/api/admin/v1/summary-stats.html)
* **Command API version 1**
    * [Commands endpoint]({{puppetdb}}/api/command/v1/commands.html)
* **Status API version 1**
    * [Status endpoint]({{puppetdb}}/api/status/v1/status.html)
* **Metadata API version 1**
    * [Version endpoint]({{puppetdb}}/api/meta/v1/version.html)
    * [Server time endpoint]({{puppetdb}}/api/meta/v1/server-time.html)
* **Metrics API version 1**
    * [Metrics endpoint]({{puppetdb}}/api/metrics/v1/mbeans.html)
    * [Changes from PuppetDB version 3]({{puppetdb}}/api/metrics/v1/changes-from-puppetdb-v3.html)
* **Wire formats**
    * [Catalog wire format - v9]({{puppetdb}}/api/wire_format/catalog_format_v9.html)
    * [Facts wire format - v5]({{puppetdb}}/api/wire_format/facts_format_v5.html)
    * [Report wire format - v8]({{puppetdb}}/api/wire_format/report_format_v8.html)
    * [Deactivate node wire format - v3]({{puppetdb}}/api/wire_format/deactivate_node_format_v3.html)

{% endmd %}
