module Puppet::Util::Puppetdb
  module CommandNames
    CommandReplaceCatalog   = "replace catalog"
    CommandReplaceFacts     = "replace facts"
    CommandDeactivateNode   = "deactivate node"
    CommandStoreReport      = "store report"
  end
end
